# Copyright (c) 2023 nggit

__all__ = ('Tremolo',)

import asyncio  # noqa: E402
import logging  # noqa: E402
import multiprocessing as mp  # noqa: E402
import os  # noqa: E402
import signal  # noqa: E402
import socket  # noqa: E402
import ssl  # noqa: E402
import sys  # noqa: E402

from functools import wraps  # noqa: E402
from importlib import import_module, reload as reload_module  # noqa: E402
from shutil import get_terminal_size  # noqa: E402

from . import __version__  # noqa: E402
from .managers import ProcessManager  # noqa: E402
from .routes import Routes  # noqa: E402
from .utils import (  # noqa: E402
    file_signature, log_date, memory_usage, server_date, getoptions
)
from .lib.connections import KeepAliveConnections  # noqa: E402
from .lib.contexts import WorkerContext  # noqa: E402
from .lib.locks import ServerLock  # noqa: E402

_REUSEPORT_OR_REUSEADDR = {
    False: socket.SO_REUSEADDR,
    True: getattr(socket, 'SO_REUSEPORT', socket.SO_REUSEADDR)
}


class Tremolo:
    def __init__(self):
        self.manager = ProcessManager()
        self.routes = Routes()
        self.middlewares = {
            'connect': [],
            'close': [],
            'request': [],
            'response': []
        }
        self.hooks = {
            'worker_start': [],
            'worker_stop': []
        }
        self.ports = {}
        self.loop = None
        self.logger = None

    def route(self, path):
        if isinstance(path, int):
            return self.error(path)

        def decorator(func):
            @wraps(func)
            def wrapper(**kwargs):
                return func(**kwargs)

            self.routes.add(wrapper, path, getoptions(func))
            return wrapper

        return decorator

    def error(self, code):
        def decorator(func):
            @wraps(func)
            def wrapper(**kwargs):
                return func(**kwargs)

            for i, h in enumerate(self.routes[0]):
                if code == h[0]:
                    self.routes[0][i] = (
                        h[0], wrapper, dict(h[2], **getoptions(func))
                    )
                    break

            return wrapper

        return decorator

    def hook(self, name, *args, priority=999):
        def decorator(func):
            @wraps(func)
            def wrapper(**kwargs):
                return func(**kwargs)

            self.add_hook(wrapper, name, priority)
            return wrapper

        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])

        return decorator

    def on_worker_start(self, *args, **kwargs):
        return self.hook('worker_start', *args, **kwargs)

    def on_worker_stop(self, *args, **kwargs):
        return self.hook('worker_stop', *args, **kwargs)

    def middleware(self, name, *args, priority=999):
        def decorator(func):
            @wraps(func)
            def wrapper(**kwargs):
                return func(**kwargs)

            self.add_middleware(wrapper, name, priority, getoptions(func))
            return wrapper

        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])

        return decorator

    def on_connect(self, *args, **kwargs):
        return self.middleware('connect', *args, **kwargs)

    def on_close(self, *args, **kwargs):
        return self.middleware('close', *args, **kwargs)

    def on_request(self, *args, **kwargs):
        return self.middleware('request', *args, **kwargs)

    def on_response(self, *args, **kwargs):
        return self.middleware('response', *args, **kwargs)

    def add_hook(self, func, name='worker_start', priority=999):
        if name not in self.hooks:
            raise ValueError('%s is not one of the: %s' %
                             (name, ', '.join(self.hooks)))

        self.hooks[name].append((priority, func))
        self.hooks[name].sort(key=lambda item: item[0],
                              reverse=name == 'worker_stop')

    def add_middleware(self, func, name='request', priority=999, kwargs=None):
        if name not in self.middlewares:
            raise ValueError('%s is not one of the: %s' %
                             (name, ', '.join(self.middlewares)))

        self.middlewares[name].append(
            (priority, func, kwargs or getoptions(func))
        )
        self.middlewares[name].sort(key=lambda item: item[0],
                                    reverse=name in ('close', 'response'))

    def listen(self, port, host=None, **options):
        if not isinstance(port, int):
            # assume it's a UNIX socket path
            host = port
            port = None

        if (host, port) in self.ports:
            return False

        self.ports[(host, port)] = options
        return (host, port) in self.ports

    async def _serve(self, host, port, **options):
        backlog = options.get('backlog', 100)

        if hasattr(options['_sock'], 'share'):
            # Windows
            sock = socket.fromshare(options['_sock'].share(os.getpid()))
        else:
            # Linux
            sock = self.create_sock(host, port, options['reuse_port'])

        sock.listen(backlog)

        if 'ssl' in options and isinstance(options['ssl'] or None, dict):
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(
                certfile=options['ssl'].get('cert', ''),
                keyfile=options['ssl'].get('key'),
                password=options['ssl'].get('password')
            )
        else:
            ssl_context = None

        server_name = options.get('server_name', b'Tremolo')

        if isinstance(server_name, str):
            server_name = server_name.encode('latin-1')

        if isinstance(host, str):
            host = host.encode('latin-1')

        lock = ServerLock(options['_locks'], loop=self.loop)
        connections = KeepAliveConnections(
            maxlen=options.get('keepalive_connections', 512)
        )
        context = WorkerContext()
        context.update(connections=connections)
        context.options.update(options)

        if options['app'] is None:
            from .http_server import HTTPServer as Server

            options['_routes'].compile()

            for _, func in self.hooks['worker_start']:
                if await func(globals=context,
                              context=context,
                              app=self,
                              loop=self.loop,
                              logger=self.logger):
                    break
        else:
            from .asgi_lifespan import ASGILifespan
            from .asgi_server import ASGIServer as Server

            # 'module:app'               -> 'module:app'   (dir: os.getcwd())
            # '/path/to/module.py'       -> 'module:app'   (dir: '/path/to')
            # '/path/to/module.py:myapp' -> 'module:myapp' (dir: '/path/to')

            if options['app'].find(':', options['app'].find(':\\') + 1) == -1:
                options['app'] += ':app'

            path, attr_name = options['app'].rsplit(':', 1)
            context.options['app_dir'], base_name = os.path.split(
                os.path.abspath(path)
            )
            module_name = os.path.splitext(base_name)[0]

            if context.options['app_dir'] == '':
                context.options['app_dir'] = os.getcwd()

            sys.path.insert(0, context.options['app_dir'])
            options['app'] = getattr(import_module(module_name), attr_name)

            print(log_date(), end=' ')
            sys.stdout.flush()
            sys.stdout.buffer.write(
                b'Starting %s as an ASGI server for: ' % server_name
            )
            print(
                getattr(options['app'], '__name__',
                        options['app'].__class__.__name__)
            )

            if server_name != b'':
                server_name += b' (ASGI)'

            context.options['_lifespan'] = ASGILifespan(
                options['app'], loop=self.loop, logger=self.logger
            )
            context.options['_lifespan'].startup()
            exc = await context.options['_lifespan'].exception(
                timeout=context.options['shutdown_timeout'] / 2
            )

            if exc:
                raise exc

        sockname = sock.getsockname()

        if isinstance(sockname, tuple):
            context.info['server'] = sockname[:2]
        else:
            context.info['server'] = (sockname, None)

        context.info['server_date'] = server_date()
        context.info['server_name'] = server_name

        server = await self.loop.create_server(
            lambda: Server(context,
                           loop=self.loop,
                           logger=self.logger,
                           lock=lock,
                           debug=options.get('debug', False),
                           ws=options.get('ws', True),
                           ws_max_payload_size=options.get(
                               'ws_max_payload_size', 2 * 1048576
                           ),
                           download_rate=options.get('download_rate', 1048576),
                           upload_rate=options.get('upload_rate', 1048576),
                           buffer_size=options.get('buffer_size', 16384),
                           client_max_body_size=options.get(
                               'client_max_body_size', 2 * 1048576
                           ),
                           client_max_header_size=options.get(
                               'client_max_header_size', 8192
                           ),
                           max_queue_size=options.get('max_queue_size', 128),
                           request_timeout=options.get('request_timeout', 30),
                           keepalive_timeout=options.get(
                               'keepalive_timeout', 30
                           ),
                           app_handler_timeout=options.get(
                               'app_handler_timeout', 120
                           ),
                           app_close_timeout=options.get(
                               'app_close_timeout', 30
                           ),
                           _app=options['app'],
                           _root_path=options.get('root_path', ''),
                           _routes=options['_routes'],
                           _middlewares=options['_middlewares']),
            sock=sock, backlog=backlog, ssl=ssl_context)

        print(log_date(), end=' ')
        sys.stdout.flush()
        sys.stdout.buffer.write(
            b'%s worker (pid %d) is started at ' % (server_name, os.getpid())
        )
        print(context.info['server'][0], end='')

        if context.info['server'][1] is not None:
            print(' port %d' % context.info['server'][1], end='')

        if ssl_context is not None:
            sys.stdout.flush()
            sys.stdout.buffer.write(b' (https)')

        print()

        try:
            await self._serve_forever(context)
        except asyncio.CancelledError:
            self.logger.info('Shutting down')

            if context.tasks:
                _, pending = await asyncio.wait(
                    context.tasks,
                    timeout=context.options['shutdown_timeout'] / 2
                )
                for task in pending:
                    task.cancel()
        finally:
            server.close()

            while context.tasks:
                await context.tasks.pop()

            await server.wait_closed()
            await self._worker_stop(context)

    async def _serve_forever(self, context):
        limit_memory = context.options.get('limit_memory', 0)
        paths = [path for path in sys.path
                 if not context.options['app_dir'].startswith(path)]
        modules = {}

        while True:
            await asyncio.sleep(1)

            # update server date
            context.info['server_date'] = server_date()

            # detect code changes
            if 'reload' in context.options and context.options['reload']:
                for module in (dict(modules) or sys.modules.values()):
                    module_file = getattr(module, '__file__', None)

                    if module_file is None:
                        continue

                    for path in paths:
                        if module_file.startswith(path):
                            break
                    else:
                        if not os.path.exists(module_file):
                            if module in modules:
                                del modules[module]

                            continue

                        sign = file_signature(module_file)

                        if module in modules:
                            if modules[module] == sign:
                                # file not modified
                                continue

                            modules[module] = sign
                        else:
                            modules[module] = sign
                            continue

                        self.logger.info('reload: %s', module_file)
                        sys.exit(3)

            if limit_memory > 0 and memory_usage() > limit_memory:
                while context.tasks:
                    context.tasks.pop().cancel()

                self.logger.error('memory limit exceeded')
                sys.exit(1)

    async def _worker_stop(self, context):
        if context.options['app'] is None:
            i = len(self.hooks['worker_stop'])

            while i > 0:
                i -= 1

                if await self.hooks['worker_stop'][i][1](
                        globals=context,
                        context=context,
                        app=self,
                        loop=self.loop,
                        logger=self.logger):
                    break
        else:
            context.options['_lifespan'].shutdown()
            exc = await context.options['_lifespan'].exception(
                timeout=context.options['shutdown_timeout'] / 2
            )

            if exc:
                self.logger.error(exc)

    def _worker(self, host, port, **kwargs):
        self.logger = logging.getLogger(mp.current_process().name)
        self.logger.setLevel(
            getattr(logging, kwargs['log_level'], logging.DEBUG)
        )

        formatter = logging.Formatter(
            kwargs.get('log_fmt',
                       '[%(asctime)s] %(module)s: %(levelname)s: %(message)s')
        )
        handler = logging.StreamHandler()

        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        loop_name = kwargs.get('loop', 'asyncio.')

        if '.' not in loop_name:
            loop_name += '.'

        # 'asyncio', '.', ''
        # 'asyncio', '.', 'SelectorEventLoop'
        module_name, _, class_name = loop_name.rpartition('.')
        module = import_module(module_name or 'asyncio')
        self.loop = getattr(module, class_name or 'new_event_loop')()

        asyncio.set_event_loop(self.loop)
        task = self.loop.create_task(self._serve(host, port, **kwargs))

        task.add_done_callback(lambda fut: self.loop.stop())
        signal.signal(signal.SIGINT, lambda signum, frame: task.cancel())
        signal.signal(signal.SIGTERM, lambda signum, frame: task.cancel())

        try:
            self.loop.run_forever()  # until loop.stop() is called
        finally:
            self.loop.close()

            if not task.cancelled():
                exc = task.exception()

                if exc:
                    raise exc

    def create_sock(self, host, port, reuse_port=True):
        try:
            try:
                socket.getaddrinfo(host, None)

                if ':' in host:
                    sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)

                    if host == '::' and hasattr(socket, 'IPPROTO_IPV6'):
                        # on Windows, Python versions below 3.8
                        # don't properly support dual-stack IPv4/6.
                        # https://github.com/python/cpython/issues/73701
                        sock.setsockopt(socket.IPPROTO_IPV6,
                                        socket.IPV6_V6ONLY, 0)
                else:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            except socket.gaierror:
                _host = host
                host = 'localhost'
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                host = _host
        except AttributeError:
            print('either AF_INET6 or AF_UNIX is not supported')
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        sock.setblocking(False)
        sock.set_inheritable(True)

        if sock.family.name == 'AF_UNIX':
            if not host.endswith('.sock'):
                host += '.sock'

            for _ in range(2):
                try:
                    sock.bind(host)
                    break
                except OSError:
                    if os.path.exists(host) and os.stat(host).st_size == 0:
                        os.unlink(host)
        else:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            sock.setsockopt(socket.SOL_SOCKET,
                            _REUSEPORT_OR_REUSEADDR[reuse_port], 1)
            sock.bind((host, port))

        return sock

    def close_sock(self, sock):
        try:
            sock.shutdown(socket.SHUT_RDWR)
            sock.close()

            if sock.family.name == 'AF_UNIX':
                os.unlink(sock.getsockname())
        except FileNotFoundError:
            pass
        except OSError:
            sock.close()

    def _handle_reload(self, **info):
        args = info['args']
        kwargs = info['kwargs']
        process = info['process']

        if process.exitcode == 3:
            print('Reloading...')

            if kwargs['app'] is None:
                for module in list(sys.modules.values()):
                    module_file = getattr(module, '__file__', None)

                    if (module_file and
                            module.__name__ not in ('__main__',
                                                    '__mp_main__',
                                                    'tremolo') and
                            not module.__name__.startswith('tremolo.') and
                            module_file.startswith(kwargs['app_dir']) and
                            os.path.exists(module_file)):
                        reload_module(module)

                module = import_module(kwargs['module_name'])

                # we need to update/rebind objects like
                # routes, middleware, etc.
                for attr in module.__dict__.values():
                    if isinstance(attr, self.__class__):
                        self.__dict__.update(attr.__dict__)

                # update some references
                kwargs['_routes'] = self.routes
                kwargs['_middlewares'] = self.middlewares
        elif process.exitcode != 0:
            print(
                'A worker process died (%d). Restarting...' % process.exitcode
            )

        if process.exitcode == 0:
            print('pid %d terminated' % process.pid)
        else:
            self.manager.spawn(
                self._worker, args=args, kwargs=kwargs,
                name=process.name,
                exit_cb=self._handle_reload
            )

    def run(self, host=None, port=0, reuse_port=True, worker_num=1, **kwargs):
        kwargs['reuse_port'] = reuse_port
        kwargs['log_level'] = kwargs.get('log_level', 'DEBUG').upper()
        kwargs.setdefault('shutdown_timeout', 30)
        server_name = kwargs.get('server_name', 'Tremolo')
        terminal_width = min(get_terminal_size()[0], 72)

        print(
            'Starting %s (tremolo %s, %s %d.%d.%d, %s)' %
            (server_name,
             __version__,
             sys.implementation.name,
             *sys.version_info[:3],
             sys.platform)
        )
        print('-' * terminal_width)

        import __main__

        if 'app' in kwargs and kwargs['app']:
            if not isinstance(kwargs['app'], str):
                if not hasattr(__main__, '__file__'):
                    raise RuntimeError('could not find ASGI app')

                for attr_name, attr in __main__.__dict__.items():
                    if attr == kwargs['app']:
                        break
                else:
                    attr_name = 'app'

                kwargs['app'] = '%s:%s' % (__main__.__file__, attr_name)

            locks = []
        else:
            kwargs['app'] = None
            locks = [mp.Lock() for _ in range(kwargs.get('locks', 16))]

            if hasattr(__main__, '__file__'):
                kwargs['app_dir'], base_name = os.path.split(
                    os.path.abspath(__main__.__file__)
                )
                kwargs['module_name'] = os.path.splitext(base_name)[0]
            else:
                kwargs['app_dir'] = os.getcwd()
                kwargs['module_name'] = '__main__'

            if kwargs['log_level'] in ('DEBUG', 'INFO'):
                print('Routes:')

                for routes in self.routes.values():
                    for route in routes:
                        pattern, func, kw = route

                        print(
                            '  %s -> %s(%s)' %
                            (pattern,
                             func.__name__,
                             ', '.join('%s=%s' % item for item in kw.items()))
                        )

                print()

        if host is None:
            if not self.ports:
                raise ValueError(
                    'with host=None, listen() must be called first'
                )

            host = 'localhost'
        else:
            self.listen(port, host=host, **kwargs)

        if worker_num < 1:
            raise ValueError('worker_num must be greater than 0')

        try:
            worker_num = min(worker_num, len(os.sched_getaffinity(0)))
        except AttributeError:
            worker_num = min(worker_num, os.cpu_count() or 1)

        socks = {}
        print('Options:')

        for (_host, _port), options in self.ports.items():
            if _host is None:
                _host = host

            if _port is None:
                _port = port

            options = {**kwargs, **options}
            print(
                '  run(host=%s, port=%d, worker_num=%d, %s)' %
                (_host,
                 _port,
                 worker_num,
                 ', '.join('%s=%s' % item for item in options.items()))
            )

            args = (_host, _port)
            socks[args] = self.create_sock(_host, _port, options['reuse_port'])

            for _ in range(options.get('worker_num', worker_num)):
                self.manager.spawn(
                    self._worker,
                    args=args,
                    kwargs=dict(options, _locks=locks, _sock=socks[args],
                                _routes=self.routes,
                                _middlewares=self.middlewares),
                    exit_cb=self._handle_reload
                )

        print('-' * terminal_width)
        print('%s main (pid %d) is running ' % (server_name, os.getpid()))

        try:
            self.manager.wait(timeout=kwargs['shutdown_timeout'])
        finally:
            for sock in socks.values():
                self.close_sock(sock)



# Copyright (c) 2023 nggit

import asyncio

from .contexts import RequestContext


class Request:
    __slots__ = ('protocol', 'context', 'body_size', 'body_consumed')

    def __init__(self, protocol):
        self.protocol = protocol
        self.context = RequestContext()
        self.body_size = 0
        self.body_consumed = 0

    @property
    def ctx(self):
        return self.context

    @property
    def transport(self):
        return self.protocol.context.transport

    @property
    def socket(self):
        return self.transport.get_extra_info('socket')

    @property
    def client(self):
        return self.protocol.context.client

    @property
    def is_secure(self):
        return bool(self.transport.get_extra_info('sslcontext'))

    def clear_body(self):
        self.body_size = 0
        self.body_consumed = 0

    async def recv(self):
        while self.protocol.queue is not None:
            task = self.protocol.loop.create_task(
                self.protocol.queue[0].get()
            )
            timer = self.protocol.loop.call_at(
                self.protocol.loop.time() +
                self.protocol.options['keepalive_timeout'],
                task.cancel
            )

            try:
                data = await task
            except asyncio.CancelledError as exc:
                raise TimeoutError('recv timeout') from exc
            finally:
                timer.cancel()

            if data is None:
                break

            self.body_consumed += len(data)
            yield data



# Copyright (c) 2023 nggit

import asyncio

from concurrent.futures import ThreadPoolExecutor


class ServerLock:
    def __init__(self, locks, name=0, timeout=30, loop=None, executors=None):
        try:
            self.name = name % len(locks)
        except ZeroDivisionError:
            return

        self.locks = locks
        self._timeout = timeout
        self._loop = loop or asyncio.get_event_loop()

        if executors is None:
            executors = {}

        self._executors = executors

        if self.name not in executors:
            self._executors[self.name] = ThreadPoolExecutor(max_workers=1)

    def __call__(self, name=0, timeout=None):
        if timeout is None:
            timeout = self._timeout

        return self.__class__(self.locks,
                              name=name,
                              timeout=timeout,
                              loop=self._loop,
                              executors=self._executors)

    async def __aenter__(self):
        try:
            await self.acquire()
        except TimeoutError:
            self.release()

    async def __aexit__(self, exc_type, exc, tb):
        self.release()

    async def acquire(self, timeout=None):
        if timeout is None:
            timeout = self._timeout

        fut = self._loop.run_in_executor(
            self._executors[self.name],
            self.locks[self.name].acquire, True, timeout)

        timer = self._loop.call_at(self._loop.time() + timeout, fut.cancel)
        result = False

        try:
            result = await fut
        except asyncio.CancelledError:
            pass
        finally:
            timer.cancel()

        if not result:
            raise TimeoutError

    def release(self):
        try:
            self.locks[self.name].release()
        except ValueError:
            pass



# Copyright (c) 2023 nggit


class Response:
    __slots__ = ('_protocol', '_send_buf')

    def __init__(self, request):
        self._protocol = request.protocol
        self._send_buf = bytearray()

    def close(self):
        if self._send_buf != b'':
            self.send_nowait(self._send_buf[:])
            del self._send_buf[:]

        self.send_nowait(None)

    async def send(self, data, rate=-1,
                   buffer_size=16384, buffer_min_size=None):
        if data is None:
            __class__.close(self)
            return

        if not isinstance(data, (bytes, bytearray)):
            raise TypeError('expected None or bytes-like object')

        if (buffer_min_size is not None and
                len(self._send_buf) < buffer_min_size):
            self._send_buf.extend(data)
        else:
            await self._protocol.put_to_queue(
                self._send_buf + data,
                i=1,
                rate=rate,
                buffer_size=buffer_size
            )

            del self._send_buf[:]

    def send_nowait(self, data):
        if self._protocol.queue is not None:
            self._protocol.queue[1].put_nowait(data)



# Copyright (c) 2023 nggit

import asyncio

from .contexts import ConnectionContext
from .http_exceptions import (
    HTTPException,
    BadRequest,
    ExpectationFailed,
    InternalServerError,
    RequestTimeout
)
from .http_header import HTTPHeader
from .http_request import HTTPRequest
from .http_response import HTTPResponse
from .queue import Queue


class HTTPProtocol(asyncio.Protocol):
    __slots__ = ('globals',
                 'context',
                 'options',
                 'loop',
                 'logger',
                 'fileno',
                 'queue',
                 'handlers',
                 'request',
                 '_watermarks',
                 '_header_buf',
                 '_waiters')

    def __init__(self, context, loop=None, logger=None, **kwargs):
        self.globals = context  # a worker-level context
        self.context = ConnectionContext()
        self.options = kwargs
        self.loop = loop
        self.logger = logger
        self.fileno = -1
        self.queue = None
        self.handlers = set()
        self.request = None

        self._watermarks = {'high': 65536, 'low': 8192}
        self._header_buf = bytearray()
        self._waiters = {}

    @property
    def transport(self):
        return self.context.transport

    @property
    def tasks(self):
        return self.context.tasks

    def add_close_callback(self, callback):
        self.tasks.add(callback)

    def create_background_task(self, coro):
        task = self.loop.create_task(coro)

        self.globals.tasks.add(task)
        task.add_done_callback(self.globals.tasks.discard)

        return task

    def create_task(self, coro):
        task = self.loop.create_task(coro)

        self.tasks.add(task)
        task.add_done_callback(self.handle_task_done)

        return task

    def handle_task_done(self, task):
        self.tasks.discard(task)

        if not task.cancelled():
            exc = task.exception()

            if exc:
                self.print_exception(exc, 'handle_task_done')

    def connection_made(self, transport):
        self.context.update(transport=transport)
        self.fileno = transport.get_extra_info('socket').fileno()

        try:
            self.queue = self.globals.queues.pop(self.fileno)
        except KeyError:
            self.queue = [Queue(), Queue()]

        self._waiters['request'] = self.loop.create_future()

        self.add_close_callback(
            self.create_background_task(self._send_data()).cancel
        )
        self.add_close_callback(self.create_background_task(
            self.set_timeout(self._waiters['request'],
                             timeout=self.options['request_timeout'],
                             timeout_cb=self.request_timeout)
        ).cancel)

    def is_closing(self):
        return self.transport is None or self.transport.is_closing()

    def close(self, exc=None):
        if self.is_closing():
            return

        if exc:
            if isinstance(exc, HTTPException):
                self.transport.write(
                    b'HTTP/1.0 %d %s\r\nContent-Type: %s\r\n\r\n' %
                    (exc.code,
                     exc.message.encode(exc.encoding),
                     exc.content_type.encode(exc.encoding))
                )
                self.transport.write(str(exc).encode(exc.encoding))

            self.print_exception(exc, 'close')

        if self.transport.can_write_eof():
            self.transport.write_eof()

        self.transport.close()

    def request_timeout(self, timeout):
        raise RequestTimeout('request timeout after %gs' % timeout)

    def keepalive_timeout(self, timeout):
        self.logger.info('keepalive timeout after %gs', timeout)

    def send_timeout(self, timeout):
        self.logger.info('send timeout after %gs', timeout)

    async def set_timeout(self, waiter, timeout=30, timeout_cb=None):
        timer = self.loop.call_at(self.loop.time() + timeout, waiter.cancel)

        try:
            return await waiter
        except asyncio.CancelledError:
            if not self.is_closing():
                try:
                    if callable(timeout_cb):
                        timeout_cb(timeout)

                    self.close()
                except Exception as exc:
                    self.close(exc)
        finally:
            timer.cancel()

    async def put_to_queue(self, data, i=0, rate=-1, buffer_size=16384):
        if data:
            mv = memoryview(data)

            while mv and self.queue is not None:
                self.queue[i].put_nowait(mv[:buffer_size].tobytes())
                queue_size = self.queue[i].qsize()

                if queue_size > self.options['max_queue_size']:
                    self.logger.error('%d exceeds the value of max_queue_size',
                                      queue_size)
                    self.close()
                    return

                if rate > 0:
                    await asyncio.sleep(1 / (rate / max(queue_size, 1) /
                                             mv[:buffer_size].nbytes))
                mv = mv[buffer_size:]
        elif self.queue is not None:
            self.queue[i].put_nowait(data)

        # maybe resume reading, or close
        if i == 0 and self.transport is not None and self.request is not None:
            if not data or self.request.upgraded:
                self.transport.resume_reading()
                return

            self.request.body_size += len(data)

            if (b'content-length' in self.request.headers and
                    self.request.body_size >= self.request.content_length and
                    self.queue is not None):
                self.queue[i].put_nowait(None)
            elif self.request.body_size < self.options['client_max_body_size']:
                if self.request.has_body:
                    self.transport.resume_reading()
            else:
                self.close(BadRequest('payload too large'))

    async def headers_received(self, response):
        raise NotImplementedError

    async def error_received(self, exc):
        raise NotImplementedError

    def handlers_timeout(self):
        if self.request is None or not self.request.upgraded:
            while self.handlers:
                self.handlers.pop().cancel()
                self.logger.error(
                    'handler timeout '
                    '(app_handler_timeout=%g, app_close_timeout=%g)',
                    self.options['app_handler_timeout'],
                    self.options['app_close_timeout']
                )

    def set_handler_timeout(self, timeout):
        if self.handlers:
            return self.loop.call_at(
                self.loop.time() + timeout, self.handlers_timeout
            )

    def print_exception(self, exc, *args):
        self.logger.error(
            ': '.join((*args, exc.__class__.__name__, str(exc))),
            exc_info=self.options['debug'] and exc
        )

    def send_continue(self):
        if self.request is None or not self.request.http_continue:
            return

        if self.request.content_length > self.options['client_max_body_size']:
            raise ExpectationFailed

        self.transport.write(
            b'HTTP/%s 100 Continue\r\n\r\n' % self.request.version
        )
        self.queue[1].put_nowait(None)

    async def _handle_request(self, header):
        self.request = HTTPRequest(self, header)
        response = HTTPResponse(self.request)

        try:
            if b'connection' in self.request.headers:
                for v in self.request.headers[b'connection'].split(b',', 100):
                    if v.strip().lower() == b'close':
                        break
                else:
                    self.globals.connections.add(self)
                    self.request.http_keepalive = True
            elif self.request.version == b'1.1':
                self.globals.connections.add(self)
                self.request.http_keepalive = True

            if self.request.has_body:
                # assuming a request with a body, such as POST
                if b'transfer-encoding' in self.request.headers:
                    if self.request.version == b'1.0':
                        raise BadRequest('unexpected chunked encoding')

                    self.request.transfer_encoding = self.request.headers[
                        b'transfer-encoding'
                    ].lower()

                if b'content-length' in self.request.headers:
                    try:
                        self.request.content_length = int(
                            b'+' + self.request.headers[b'content-length']
                        )
                    except (ValueError, TypeError) as exc:
                        raise BadRequest('bad Content-Length') from exc

                    if (b'%d' % self.request.content_length !=
                            self.request.headers[b'content-length'] or
                            b'chunked' in self.request.transfer_encoding):
                        raise BadRequest

                if (b'expect' in self.request.headers and
                        self.request.headers[b'expect']
                        .lower() == b'100-continue'):
                    # we can handle continue later after the route is found
                    # by checking this state
                    self.request.http_continue = True
            else:
                await self.put_to_queue(b'')

            if self.request.has_body or header.body:
                # the initial body that accompanies the header
                # or the next request header, if it's a bodyless request
                await self.put_to_queue(
                    header.body, rate=self.options['upload_rate']
                )

            # successfully got header,
            # clear either the request or keepalive timeout
            if not self._waiters['request'].done():
                self._waiters['request'].set_result(None)

            timer = self.set_handler_timeout(
                self.options['app_handler_timeout']
            )

            try:
                if self.request is not None:
                    await self.headers_received(response)
            finally:
                timer.cancel()
        except (asyncio.CancelledError, Exception) as exc:
            data = None

            try:
                data = await self.error_received(exc)
            finally:
                await response.handle_exception(exc, data)

    async def _receive_data(self, data, waiter):
        await waiter
        await self.put_to_queue(
            data,
            rate=self.options['upload_rate'],
            buffer_size=self.options['buffer_size']
        )

    def data_received(self, data):
        if not data:
            return

        if self._header_buf is not None:
            self._header_buf.extend(data)
            header_size = self._header_buf.find(b'\r\n\r\n') + 2

            if 1 < header_size <= self.options['client_max_header_size']:
                # this will keep blocking on bodyless requests forever, unless
                # Response.close is called (resumed in _send_data)
                self.transport.pause_reading()

                header = HTTPHeader(self._header_buf,
                                    header_size=header_size,
                                    excludes=[b'proxy'])

                if header.is_request:
                    task = self.create_background_task(
                        self._handle_request(header)
                    )
                    self.handlers.add(task)
                    task.add_done_callback(self.handlers.discard)
                else:
                    self.close(BadRequest('bad request: not a request'))

                self._header_buf = None
            elif header_size > self.options['client_max_header_size']:
                self.close(BadRequest('request header too large'))
            elif not (header_size == 1 and len(self._header_buf) <=
                      self.options['client_max_header_size']):
                self.close(BadRequest('bad request'))

            return

        # resumed in put_to_queue or _send_data
        self.transport.pause_reading()

        if 'receive' in self._waiters:
            waiter = self._waiters['receive']
        else:
            waiter = self._waiters['request']

        self._waiters['receive'] = self.create_task(
            self._receive_data(data, waiter)
        )

    def eof_received(self):
        self.queue[0].put_nowait(None)

    def resume_writing(self):
        if 'send' in self._waiters and not self._waiters['send'].done():
            self._waiters['send'].set_result(None)

    def set_watermarks(self, high=65536, low=8192):
        if self.transport is not None:
            self._watermarks['high'] = high
            self._watermarks['low'] = low

            self.transport.set_write_buffer_limits(high=high, low=low)

    async def _send_data(self):
        while self.queue is not None:
            try:
                data = await self.queue[1].get()

                if data is None:
                    # close the transport, unless keepalive is enabled
                    if self.request is not None:
                        if self.request.http_continue:
                            self.request.http_continue = False
                            self.transport.resume_reading()
                            continue

                        if self.request.http_keepalive:
                            self.request.http_keepalive = False

                            if self in self.globals.connections:
                                await self._handle_keepalive()
                                self.transport.resume_reading()
                                continue

                            self.logger.info(
                                'keepalive connection dropped: %d', self.fileno
                            )

                        self.request.clear_body()

                    self.close()
                    return

                # send data
                write_buffer_size = self.transport.get_write_buffer_size()

                if write_buffer_size > self._watermarks['high']:
                    self.logger.info(
                        '%d exceeds the current watermark limits '
                        '(high=%d, low=%d)',
                        write_buffer_size,
                        self._watermarks['high'],
                        self._watermarks['low']
                    )
                    self._waiters['send'] = self.loop.create_future()

                    await self.set_timeout(
                        self._waiters['send'],
                        timeout=self.options['keepalive_timeout'],
                        timeout_cb=self.send_timeout
                    )

                    if self.transport is None:
                        return

                self.transport.write(data)
            except asyncio.CancelledError:
                self.close()
                break
            except Exception as exc:
                self.close(exc)
                break

    async def _handle_keepalive(self):
        if self.request.upgraded:
            self._waiters.setdefault('receive', self._waiters.pop('request'))
        else:
            if 'receive' in self._waiters:
                # waits for all incoming data to enter the queue
                await self._waiters.pop('receive')

            if self.request.body_size < self.request.content_length:
                raise InternalServerError(
                    'request body was not fully consumed'
                )

            # reset. so the next data in data_received will be considered as
            # a fresh http request (not a continuation data)
            self.request.clear_body()
            self._header_buf = bytearray()

            self._waiters['request'] = self.loop.create_future()

            self.add_close_callback(self.create_background_task(
                self.set_timeout(self._waiters['request'],
                                 timeout=self.options['keepalive_timeout'],
                                 timeout_cb=self.keepalive_timeout)
            ).cancel)

            while not self.request.has_body and self.queue[0].qsize():
                # this data is supposed to be the next header
                self.data_received(self.queue[0].get_nowait())

    def connection_lost(self, _):
        while self.tasks:
            task = self.tasks.pop()

            try:
                if callable(task):
                    # a close callback
                    task()
                    continue

                task.cancel()
            except Exception as exc:
                self.print_exception(exc, 'connection_lost')

        if self.queue is not None:
            if self.queue[0].clear() and self.queue[1].clear():
                self.globals.queues[self.fileno] = self.queue

            self.queue = None

        self.context.update(transport=None)
        self.request = None
        self._header_buf = None

        self.set_handler_timeout(self.options['app_close_timeout'])
        self.globals.connections.discard(self)



# Copyright (c) 2023 nggit

from .lib.http_exceptions import (  # noqa: F401
    TremoloException,
    HTTPException,
    BadRequest,
    Unauthorized,
    Forbidden,
    NotFound,
    MethodNotAllowed,
    RequestTimeout,
    PreconditionFailed,
    PayloadTooLarge,
    RangeNotSatisfiable,
    ExpectationFailed,
    TooManyRequests,
    InternalServerError,
    ServiceUnavailable,
    WebSocketException,
    WebSocketClientClosed,
    WebSocketServerClosed
)


class ASGIException(TremoloException):
    message = 'ASGIException'


class LifespanError(ASGIException):
    pass


class LifespanProtocolUnsupported(ASGIException):
    message = 'ASGI Lifespan Protocol is not supported by your application'



# Copyright (c) 2023 nggit

from .lib.http_protocol import HTTPProtocol
from .lib.http_response import KEEPALIVE_OR_CLOSE, UPGRADE_OR_KEEPALIVE
from .lib.sse import SSE
from .lib.websocket import WebSocket


class HTTPServer(HTTPProtocol):
    __slots__ = ('_routes', '_middlewares', '_server')

    def __init__(self, context, _routes=None, _middlewares=None, **kwargs):
        super().__init__(context, **kwargs)

        self._routes = _routes
        self._middlewares = _middlewares
        self._server = {
            'response': None,  # set in headers_received
            'globals': context,
            'context': self.context,
            'loop': kwargs['loop'],
            'logger': kwargs['logger'],
            'lock': kwargs['lock']
        }

    @property
    def response(self):
        return self._server['response']

    async def _connection_made(self):
        for _, func, _ in self._middlewares['connect']:
            if await func(**self._server):
                break

    async def _connection_lost(self, exc):
        try:
            i = len(self._middlewares['close'])

            while i > 0:
                i -= 1

                if await self._middlewares['close'][i][1](**self._server):
                    break
        finally:
            super().connection_lost(exc)

    def connection_made(self, transport):
        super().connection_made(transport)

        if self._middlewares['connect']:
            self.context.ON_CONNECT = self.create_background_task(
                self._connection_made()
            )
            self.add_close_callback(self.context.ON_CONNECT.cancel)

    def connection_lost(self, exc):
        if self._middlewares['close']:
            task = self.create_background_task(self._connection_lost(exc))
            self.loop.call_at(
                self.loop.time() + self.options['app_close_timeout'],
                task.cancel
            )
        else:
            super().connection_lost(exc)

        self._server['response'] = None

    async def run_middlewares(self, name, reverse=False, start=0, step=1):
        if reverse:
            start = len(self._middlewares[name]) - 1
            step = -1

        while 0 <= start < len(self._middlewares[name]):
            if await self._handle_middleware(
                    self._middlewares[name][start][1],
                    self._middlewares[name][start][2]):
                return True

            start += step

    async def _handle_middleware(self, func, kwargs):
        options = self.request.context.options
        options.update(kwargs)

        data = await func(**self._server)

        if data is None:
            return False

        if isinstance(data, (bytes, bytearray, str, tuple)):
            if 'status' in options:
                self.response.set_status(*options['status'])

            if 'content_type' in options:
                self.response.set_content_type(options['content_type'])

            encoding = ('utf-8',)

            if isinstance(data, tuple):
                data, *encoding = data + encoding

            if isinstance(data, str):
                data = data.encode(encoding[0])

            await self.response.end(data)
        else:
            self.logger.info('middleware %s has exited with the connection '
                             'possibly left open', func.__name__)

        return True

    def _handle_websocket(self):
        if (self.options['ws'] and
                b'sec-websocket-key' in self.request.headers and
                b'upgrade' in self.request.headers and
                self.request.headers[b'upgrade'].lower() == b'websocket'):
            self._server['websocket'] = WebSocket(self.request, self.response)

    async def _handle_response(self, func, kwargs):
        options = self.request.context.options
        options.update(kwargs)
        options.setdefault('rate', self.options['download_rate'])
        options.setdefault('buffer_size', self.options['buffer_size'])

        if not self.request.has_body:
            if 'websocket' in options:
                self._handle_websocket()

            if 'sse' in options:
                self._server['sse'] = SSE(self.request, self.response)

        if 'status' in options:
            self.response.set_status(*options['status'])

        if 'content_type' in options:
            self.response.set_content_type(options['content_type'])

        try:
            agen = func(**self._server)
        except TypeError:  # doesn't accept extra **kwargs
            agen = func(**{k: self._server.get(k, kwargs[k]) for k in kwargs})

        next_data = getattr(agen, '__anext__', False)

        if next_data:
            data = await next_data()
        else:
            data = await agen

            if data is None:
                self.response.close()
                return

            if not isinstance(data, (bytes, bytearray, str, tuple)):
                self.logger.info('handler %s has exited with the connection '
                                 'possibly left open', func.__name__)
                return

        status = self.response.get_status()
        no_content = status[0] in (204, 205, 304) or 100 <= status[0] < 200
        self.response.http_chunked = options.get(
            'chunked', self.request.version == b'1.1' and not no_content
        )
        self.response.set_status(*status)

        if not no_content:
            self.response.set_header(
                b'Content-Type', self.response.get_content_type()
            )

        if self.response.http_chunked:
            self.response.set_header(b'Transfer-Encoding', b'chunked')

        if next_data:
            if self.request.http_keepalive:
                if status[0] == 101:
                    self.request.upgraded = True
                elif not (self.response.http_chunked or
                          b'content-length' in self.response.headers):
                    # no chunk, no close, no size.
                    # Assume close to signal end
                    self.request.http_keepalive = False

                self.response.set_header(
                    b'Connection',
                    UPGRADE_OR_KEEPALIVE[status[0] in (101, 426)]
                )
            else:
                self.response.set_header(b'Connection', b'close')

            if await self.run_middlewares('response', reverse=True):
                return

            if self.request.method == b'HEAD' or no_content:
                await self.response.write(None)
                return

            buffer_min_size = options['buffer_size'] // 2

            self.set_watermarks(high=options['buffer_size'] * 4,
                                low=buffer_min_size)

            if options.get('stream', True):
                buffer_min_size = None

            if data != b'':
                await self.response.write(data,
                                          rate=options['rate'],
                                          buffer_size=options['buffer_size'],
                                          buffer_min_size=buffer_min_size)

            while True:
                try:
                    data = await next_data()

                    if data == b'':
                        continue

                    await self.response.write(
                        data,
                        rate=options['rate'],
                        buffer_size=options['buffer_size'],
                        buffer_min_size=buffer_min_size
                    )
                except StopAsyncIteration:
                    await self.response.write(
                        b'', buffer_min_size=buffer_min_size
                    )
                    break
        else:
            encoding = ('utf-8',)

            if isinstance(data, tuple):
                data, *encoding = data + encoding

            if isinstance(data, str):
                data = data.encode(encoding[0])

            if not (no_content or self.response.http_chunked):
                self.response.set_header(b'Content-Length', b'%d' % len(data))

            self.response.set_header(
                b'Connection', KEEPALIVE_OR_CLOSE[self.request.http_keepalive]
            )

            if await self.run_middlewares('response', reverse=True):
                return

            if self.request.method == b'HEAD' or no_content:
                await self.response.write(None)
                return

            if data != b'':
                self.set_watermarks(high=options['buffer_size'] * 4,
                                    low=options['buffer_size'] // 2)
                await self.response.write(data,
                                          rate=options['rate'],
                                          buffer_size=options['buffer_size'])

            await self.response.write(b'')

        self.response.close(keepalive=True)

    async def headers_received(self, response):
        self._server['response'] = response
        self._server['request'] = self.request

        if self._middlewares['connect']:
            await self.context.ON_CONNECT

        if await self.run_middlewares('request'):
            return

        if not self.request.is_valid:
            # bad request
            await self._handle_response(
                self._routes[0][0][1], self._routes[0][0][2]
            )
            return

        path = self.request.path.strip(b'/')

        if path == b'':
            key = 1
        else:
            parts = path.split(b'/', 254)
            key = bytes([len(parts)]) + parts[0]

        if key in self._routes:
            for (pattern, func, kwargs) in self._routes[key]:
                m = pattern.search(self.request.url)

                if m:
                    matches = m.groupdict()

                    if not matches:
                        matches = m.groups()

                    self.request.params['path'] = matches

                    await self._handle_response(func, kwargs)
                    return
        else:
            i = len(self._routes[-1])

            while i > 0:
                i -= 1
                pattern, func, kwargs = self._routes[-1][i]
                m = pattern.search(self.request.url)

                if m:
                    if key in self._routes:
                        self._routes[key].append((pattern, func, kwargs))
                    else:
                        self._routes[key] = [(pattern, func, kwargs)]

                    matches = m.groupdict()

                    if not matches:
                        matches = m.groups()

                    self.request.params['path'] = matches

                    await self._handle_response(func, kwargs)
                    del self._routes[-1][i]
                    return

        # not found
        await self._handle_response(
            self._routes[0][1][1], self._routes[0][1][2]
        )

    async def error_received(self, exc):
        # internal server error
        if self.request is not None and self.response is not None:
            return await self._routes[0][-1][1](request=self.request,
                                                response=self.response,
                                                exc=exc)



# Copyright (c) 2023 nggit

import multiprocessing as mp
import os
import signal

from threading import Thread

PARENT = 0
CHILD = 1


def sigterm_handler(signum, frame):
    raise KeyboardInterrupt


class ProcessManager:
    processes = {}

    @classmethod
    def _wait_main(cls, conn):
        while True:
            try:
                if conn.poll(1):
                    conn.recv()
            except EOFError:  # parent has exited
                os._exit(0)
                break
            except OSError:  # handle is closed
                break

    @classmethod
    def _target(cls, conn, func, *args, **kwargs):
        conn[PARENT].close()

        t = Thread(target=cls._wait_main, args=(conn[CHILD],))
        t.start()

        try:
            conn[CHILD].send(os.getpid())  # started, send pid to parent

            func(*args, **kwargs)
        except KeyboardInterrupt:
            pass
        finally:
            conn[CHILD].close()  # trigger handle is closed, also notify parent
            t.join()

    def spawn(self, target, args=(), kwargs={}, name=None, exit_cb=None):
        conn = mp.Pipe()
        process = mp.Process(target=self._target, name=name,
                             args=(conn, target, *args), kwargs=kwargs)
        process.start()
        conn[CHILD].close()

        self.processes[process.name] = {
            'target': target,
            'args': args,
            'kwargs': kwargs,
            'exit_cb': exit_cb,
            'parent_conn': conn[PARENT],
            'process': process
        }
        # block until the child starts, receive its pid
        return conn[PARENT].recv()

    def wait(self, timeout=30):
        signal.signal(signal.SIGTERM, sigterm_handler)

        while self.processes:
            try:
                connections = [info['parent_conn'] for info in
                               self.processes.values()]
                for conn in mp.connection.wait(connections):
                    # a child has exited, clean up
                    # there is no need to call recv() since EOF is expected
                    for name in self.processes:
                        if self.processes[name]['parent_conn'] is conn:
                            info = self.processes.pop(name)

                            info['process'].join()
                            conn.close()

                            if callable(info['exit_cb']):
                                info['exit_cb'](**info)

                            break
            except KeyboardInterrupt:
                while self.processes:
                    _, info = self.processes.popitem()
                    process = info['process']

                    try:
                        os.kill(process.pid, signal.SIGINT)
                        process.join(timeout)
                    except (OSError, KeyboardInterrupt):
                        print('pid %d terminated (forced quit)' % process.pid)
                    finally:
                        info['parent_conn'].close()

                    if process.exitcode == 0 and callable(info['exit_cb']):
                        info['exit_cb'](**info)



# Copyright (c) 2023 nggit

from traceback import TracebackException

from .exceptions import BadRequest
from .utils import html_escape


async def index(**_):
    return b'Service Unavailable'


async def error_400(**_):
    raise BadRequest


async def error_404(request, globals, **_):
    yield (
        b'<!DOCTYPE html><html lang="en"><head><meta name="viewport" '
        b'content="width=device-width, initial-scale=1.0" />'
        b'<title>404 Not Found</title>'
        b'<style>body { max-width: 600px; margin: 0 auto; padding: 1%; '
        b'font-family: sans-serif; line-height: 1.5em; }</style></head>'
        b'<body><h1>Not Found</h1>'
    )
    yield (b'<p>Unable to find handler for %s.</p><hr />' %
           html_escape(request.path))
    yield (
        b'<address title="Powered by Tremolo">%s</address>'
        b'</body></html>' % globals.info['server_name']
    )


async def error_500(request, exc=None, **_):
    if request.protocol.options['debug']:
        te = TracebackException.from_exception(exc)
        return '<ul><li>%s</li></ul>' % '</li><li>'.join(
            html_escape(line) for line in te.format()
        )

    return str(exc)



# Copyright (c) 2023 nggit

import asyncio

from collections import deque


class Queue:
    def __init__(self, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._queue = deque()
        self._getters = deque()

    def qsize(self):
        return len(self._queue)

    def put_nowait(self, item):
        while self._getters:
            fut = self._getters.popleft()

            if not fut.done():
                fut.set_result(item)
                return

        self._queue.append(item)

    async def get(self):
        if self._getters:
            await self._getters[-1]

        try:
            return self._queue.popleft()
        except IndexError:
            fut = self._loop.create_future()
            self._getters.append(fut)

            return await fut

    def get_nowait(self):
        return self._queue.popleft()

    # non-standar
    def clear(self):
        self._queue.clear()
        self._getters.clear()

        return not self._queue and not self._getters



# Copyright (c) 2023 nggit

import asyncio

from http import HTTPStatus
from urllib.parse import unquote_to_bytes

from .exceptions import (
    InternalServerError,
    WebSocketClientClosed,
    WebSocketServerClosed
)
from .handlers import error_400, error_500
from .lib.http_protocol import HTTPProtocol
from .lib.websocket import WebSocket

_WSS_OR_WS = {
    False: 'ws',
    True: 'wss'
}


class ASGIServer(HTTPProtocol):
    __slots__ = ('response', '_scope', '_read', '_websocket', '_waiter')

    def __init__(self, context, **kwargs):
        super().__init__(context, **kwargs)

        self._scope = {
            'asgi': {'version': '3.0', 'spec_version': '2.3'},
            'root_path': self.options['_root_path'],
            'server': self.globals.info['server']
        }
        self.response = None  # set in headers_received
        self._read = None
        self._websocket = None
        self._waiter = self.loop.create_future()

    def _handle_websocket(self):
        self._websocket = WebSocket(self.request, self.response)

        self._scope['type'] = 'websocket'
        self._scope['scheme'] = _WSS_OR_WS[self.request.scheme == b'https']
        self._scope['subprotocols'] = [
            value.decode('latin-1') for value in
            self.request.headers.getlist(b'sec-websocket-protocol')]

    def _handle_http(self):
        self._read = self.request.stream()

        self._scope['type'] = 'http'
        self._scope['method'] = self.request.method.decode('latin-1')
        self._scope['scheme'] = self.request.scheme.decode('latin-1')

    async def headers_received(self, response):
        self.response = response

        if not self.request.is_valid:
            await error_400(request=self.request, response=response)
            return

        self._scope.update(
            http_version=self.request.version.decode('latin-1'),
            path=unquote_to_bytes(self.request.path).decode('latin-1'),
            raw_path=self.request.path,
            query_string=self.request.query_string,
            headers=self.request.header.getheaders(),
            client=self.request.client
        )

        if (self.options['ws'] and
                b'sec-websocket-key' in self.request.headers and
                b'upgrade' in self.request.headers and
                self.request.headers[b'upgrade'].lower() == b'websocket'):
            self._handle_websocket()
        else:
            self._handle_http()

        try:
            await self.options['_app'](self._scope, self.receive, self.send)

            if not self._waiter.done():
                self.logger.info('handler exited early (no close?)')
                await self._waiter
        except (asyncio.CancelledError, Exception) as exc:
            if (self.request is not None and self.request.upgraded and
                    self._websocket is not None):
                exc = WebSocketServerClosed(cause=exc)

            await response.handle_exception(exc)

    async def error_received(self, exc):
        if self.request is not None and self.response is not None:
            return await error_500(
                request=self.request, response=self.response, exc=exc
            )

    def connection_lost(self, exc):
        super().connection_lost(exc)

        if not self._waiter.done():
            self._waiter.set_result(None)

        self.response = None
        self._scope = None

    async def receive(self):
        if self._scope['type'] == 'websocket':
            # initially, the Request.upgraded value is False
            # it will become True later
            # after the response status is set to 101:
            # Response.set_status(101) in WebSocket.accept()
            if not self.request.upgraded:
                return {'type': 'websocket.connect'}

            try:
                payload = await self._websocket.receive()

                if isinstance(payload, str):
                    return {
                        'type': 'websocket.receive',
                        'text': payload
                    }

                return {
                    'type': 'websocket.receive',
                    'bytes': payload
                }
            except (asyncio.CancelledError, Exception) as exc:
                code = 1005

                if isinstance(exc, WebSocketClientClosed):
                    code = exc.code

                if self._websocket is None:
                    self.logger.info(
                        'calling receive() after the connection is closed'
                    )
                else:
                    self.print_exception(exc)

                self.set_handler_timeout(self.options['app_close_timeout'])
                return {
                    'type': 'websocket.disconnect',
                    'code': code
                }

        if self._scope['type'] != 'http':
            await self.response.handle_exception(
                InternalServerError('unsupported scope type %s' %
                                    self._scope['type'])
            )
            return

        try:
            data = await self._read.__anext__()
            more_body = (
                (data != b'' and self.request.content_length == -1) or
                self.request.body_consumed < self.request.content_length
            )

            if not more_body:
                self._read = None

            return {
                'type': 'http.request',
                'body': data,
                'more_body': more_body
            }
        except (asyncio.CancelledError, Exception) as exc:
            if self.request is None:
                self.logger.info(
                    'calling receive() after the connection is closed'
                )
            elif self._read is None or isinstance(exc, StopAsyncIteration):
                # delay http.disconnect (a workaround for Quart)
                # https://github.com/nggit/tremolo/issues/202
                if self._waiter.done():
                    self._waiter = self.loop.create_future()

                if self._read is not None:
                    self._read = None
                    return {'type': 'http.request'}

                self.logger.info(
                    'calling receive() when there is no more body'
                )
                await self._waiter
            else:
                self.print_exception(exc)

            self.set_handler_timeout(self.options['app_close_timeout'])
            return {'type': 'http.disconnect'}

    async def send(self, data):
        try:
            if data['type'] in ('http.response.start', 'websocket.accept'):
                # websocket doesn't have this
                if 'status' in data:
                    self.response.set_status(data['status'],
                                             HTTPStatus(data['status']).phrase)

                if 'headers' in data:
                    for header in data['headers']:
                        if b'\n' in header[0] or b'\n' in header[1]:
                            await self.response.handle_exception(
                                InternalServerError(
                                    'name or value cannot contain '
                                    'illegal characters')
                            )
                            return

                        name = header[0].lower()

                        if name == b'content-type':
                            self.response.set_content_type(header[1])
                            continue

                        if name in (b'date', b'server', b'transfer-encoding'):
                            # disallow apps from changing them,
                            # as they are managed by Tremolo
                            continue

                        if name == b'connection':
                            if header[1].lower() == b'close':
                                self.request.http_keepalive = False

                            continue

                        if name == b'content-length':
                            # will disable http chunked in the
                            # self.response.write()
                            self.response.http_chunked = False

                        self.response.append_header(*header)

                # websocket has this
                if 'subprotocol' in data and data['subprotocol']:
                    if '\n' in data['subprotocol']:
                        await self.response.handle_exception(
                            InternalServerError(
                                'subprotocol value cannot contain '
                                'illegal characters')
                        )
                        return

                    self.response.set_header(
                        b'Sec-WebSocket-Protocol',
                        data['subprotocol'].encode('utf-8')
                    )

            if data['type'] == 'http.response.body':
                if 'body' in data and data['body'] != b'':
                    await self.response.write(
                        data['body'],
                        rate=self.options['download_rate'],
                        buffer_size=self.options['buffer_size']
                    )

                if 'more_body' not in data or data['more_body'] is False:
                    await self.response.write(b'')
                    self.response.close(keepalive=True)
                    self._read = None  # disallows further receive()
                    self._waiter.cancel()
            elif data['type'] == 'websocket.send':
                if 'bytes' in data and data['bytes']:
                    await self._websocket.send(data['bytes'])
                elif 'text' in data and data['text']:
                    await self._websocket.send(data['text'], opcode=1)
            elif data['type'] == 'websocket.accept':
                await self._websocket.accept()
            elif data['type'] == 'websocket.close':
                await self._websocket.close(data.get('code', 1000))
                self._websocket = None
                self._waiter.cancel()
        except asyncio.CancelledError:
            pass
        except Exception as exc:
            if self.request is None or self.response is None:
                self.logger.info(
                    'calling send() after the connection is closed'
                )
            else:
                self.print_exception(exc)



# Copyright (c) 2023 nggit


class Headers(dict):
    def copy(self):
        return self.__class__(self)

    def getlist(self, name):
        values = self.get(name, [])

        if isinstance(values, list):
            result = []

            for v in values:
                result.extend(v.replace(b', ', b',').split(b',', 100))

            return result

        return values.replace(b', ', b',').split(b',', 100)


class HTTPHeader:
    __slots__ = ('is_request', 'is_response', 'is_valid', 'headers',
                 '_headers', 'body')

    def __init__(self, data=None, **kwargs):
        self.is_request = False
        self.is_response = False
        self.is_valid = False

        self.headers = Headers()
        self._headers = []
        self.body = b''

        self.parse(data, **kwargs)

    def parse(self, data, header_size=-1, header_max_size=65536, excludes=(),
              max_lines=100, max_line_size=8190):
        if not data:
            return self

        if header_size == -1:
            header_size = data.find(b'\r\n\r\n', 0, header_max_size) + 2

        if header_size < 2:
            return self

        self.is_request = False
        self.is_response = False
        self.is_valid = False

        self.headers.clear()
        self._headers.clear()
        self.body = data[header_size + 2:]  # store data after the header
        start = 0

        while True:
            end = data.find(b'\r\n', start, header_size)

            if end == -1:
                break

            max_lines -= 1
            line = bytes(data[start:end])

            if (max_lines < 0 or end - start > max_line_size or
                    b'\r' in line or b'\n' in line or 0 in line):
                self.is_valid = False
                return self

            colon_pos = line.find(b':', 1)

            if start == 0:
                http_pos = line.rfind(b'HTTP/')

                if http_pos == 0:
                    self.is_response = True

                    try:
                        (
                            _,
                            self.headers[b'_version'],
                            _status,
                            self.headers[b'_message']
                        ) = line.replace(b'/', b' ').split(b' ', 3)
                        self.headers[b'_status'] = int(_status)
                        self.is_valid = True
                    except ValueError:
                        self.headers[b'_version'] = b''
                        self.headers[b'_status'] = 0
                        self.headers[b'_message'] = b''
                elif http_pos > 1 and line[http_pos - 1] == 32:
                    self.is_request = True

                    try:
                        (
                            self.headers[b'_method'],
                            self.headers[b'_url']
                        ) = line[:http_pos - 1].split(b' ', 1)
                        self.headers[b'_version'] = line[http_pos + 5:]
                        self.is_valid = True
                    except ValueError:
                        self.headers[b'_method'] = b''
                        self.headers[b'_url'] = b''
                        self.headers[b'_version'] = b''

                self.headers[b'_line'] = line
            elif colon_pos > 0 and line[colon_pos - 1] != 32:
                name = line[:colon_pos].lower()
                value = line[colon_pos + 1:].strip(b' \t')

                if name in self.headers:
                    if isinstance(self.headers[name], list):
                        self.headers[name].append(value)
                    else:
                        self.headers[name] = [self.headers[name], value]
                else:
                    self.headers[name] = value

                if name not in excludes:
                    self._headers.append((name, value))
            else:
                self.is_valid = False
                break

            start = end + 2

        if self.is_request and b'host' not in self.headers:
            self.headers[b'host'] = b''

            if self.is_valid and self.headers[b'_version'] == b'1.1':
                self.is_valid = False

        return self

    def remove(self, *args):
        if not args:
            return self

        i = len(self._headers)

        while i > 0:
            i -= 1

            if self._headers[i][0] in args:
                del self._headers[i]

        return self

    def append(self, *args):
        for v in args:
            if isinstance(v, tuple):
                self._headers.append(v)

        return self

    def getheaders(self):
        return self._headers

    def gethost(self):
        return self.headers.get(b'x-forwarded-host', self.headers.get(b'host'))

    def getmethod(self):
        return self.headers.get(b'_method')

    def geturl(self):
        return self.headers.get(b'_url')

    def getversion(self):
        return self.headers.get(b'_version')

    def getstatus(self):
        return self.headers.get(b'_status')

    def getmessage(self):
        return self.headers.get(b'_message')

    def save(self):
        return b'\r\n'.join(
            [self.headers.get(b'_line', b'')] +
            [b': '.join(v) for v in self._headers]
        ) + b'\r\n\r\n' + self.body



__version__ = '0.1.5'

from .tremolo import Tremolo, Tremolo as Application  # noqa: E402,F401
from . import exceptions  # noqa: E402,F401


def run(app, **options):
    if 'host' not in options:
        options['host'] = '127.0.0.1'

    if 'port' not in options:
        options['port'] = 8000

    Tremolo().run(app=app, **options)



# Copyright (c) 2023 nggit

import re

from . import handlers
from .utils import getoptions


class Routes(dict):
    def __init__(self):
        self[0] = [
            (400, handlers.error_400, {}),
            (404, handlers.error_404, dict(request=None,
                                           globals=None,
                                           status=(404, b'Not Found'),
                                           stream=False)),
            # must be at the very end
            (500, handlers.error_500, {})
        ]
        self[1] = [
            (
                b'^/+(?:\\?.*)?$',
                handlers.index, dict(status=(503, b'Service Unavailable'))
            )
        ]
        self[-1] = []

    def add(self, func, path='/', kwargs=None):
        if not kwargs:
            kwargs = getoptions(func)

        if path.startswith('^') or path.endswith('$'):
            pattern = path.encode('latin-1')
            self[-1].append((pattern, func, kwargs))
        else:
            path = path.split('?', 1)[0].strip('/').encode('latin-1')

            if path == b'':
                key = 1
                pattern = self[1][0][0]
                self[key] = [(pattern, func, kwargs)]
            else:
                parts = path.split(b'/', 254)
                key = bytes([len(parts)]) + parts[0]
                pattern = b'^/+%s(?:/+)?(?:\\?.*)?$' % path

                if key in self:
                    self[key].append((pattern, func, kwargs))
                else:
                    self[key] = [(pattern, func, kwargs)]

    def compile(self):
        for key in self:
            for i, h in enumerate(self[key]):
                pattern, *handler = h

                if isinstance(pattern, bytes):
                    self[key][i] = (re.compile(pattern), *handler)



# Copyright (c) 2023 nggit


class Context:
    def __init__(self):
        self.__dict__ = {}

    def __repr__(self):
        return self.__dict__.__repr__()

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def __setitem__(self, *args):
        self.__setattr__(*args)

    def __getitem__(self, name):
        return self.__dict__[name]

    def __contains__(self, name):
        return self.__dict__.__contains__(name)


class WorkerContext(Context):
    def __init__(self):
        self.__dict__ = {
            'info': {},
            'options': {},
            'queues': {},
            'tasks': set()
        }

    @property
    def info(self):
        return self.__dict__['info']

    @property
    def options(self):
        return self.__dict__['options']

    @property
    def queues(self):
        return self.__dict__['queues']

    @property
    def tasks(self):
        return self.__dict__['tasks']

    @property
    def connections(self):
        return self.__dict__.get('connections', None)


class ConnectionContext(Context):
    def __init__(self):
        self.__dict__ = {
            'transport': None,
            'tasks': set()
        }

    @property
    def transport(self):
        return self.__dict__['transport']

    @property
    def client(self):
        if 'client' not in self.__dict__ and self.transport is not None:
            sock = self.transport.get_extra_info('socket')
            self.__dict__['client'] = sock.getpeername() or None

            if isinstance(self.__dict__['client'], tuple):
                self.__dict__['client'] = self.__dict__['client'][:2]

        return self.__dict__.get('client', None)

    @property
    def tasks(self):
        return self.__dict__['tasks']


class RequestContext(Context):
    def __init__(self):
        self.__dict__ = {
            'options': {}
        }

    @property
    def options(self):
        return self.__dict__['options']



# Copyright (c) 2023 nggit

import os
import time

from urllib.parse import parse_qs

from tremolo.utils import parse_fields
from .http_exceptions import BadRequest, PayloadTooLarge
from .request import Request


class HTTPRequest(Request):
    __slots__ = ('_ip',
                 '_scheme',
                 'header',
                 'headers',
                 'is_valid',
                 'host',
                 'method',
                 'url',
                 'path',
                 'query_string',
                 'version',
                 'content_length',
                 'transfer_encoding',
                 'http_continue',
                 'http_keepalive',
                 '_upgraded',
                 '_body',
                 '_eof',
                 '_read_instance',
                 '_read_buf')

    def __init__(self, protocol, header):
        super().__init__(protocol)

        self._ip = None
        self._scheme = None
        self.header = header
        self.headers = header.headers.copy()
        self.is_valid = header.is_valid
        self.host = header.gethost()

        if isinstance(self.host, list):
            self.host = self.host[0]

        self.method = header.getmethod().upper()
        self.url = header.geturl()
        self.path, _, self.query_string = self.url.partition(b'?')
        self.version = header.getversion()

        if self.version != b'1.0':
            self.version = b'1.1'

        self.content_length = -1
        self.transfer_encoding = b'none'
        self.http_continue = False
        self.http_keepalive = False
        self._upgraded = False
        self._body = bytearray()
        self._eof = False
        self._read_instance = None
        self._read_buf = bytearray()

    @property
    def ip(self):
        if not self._ip:
            ip = self.headers.get(b'x-forwarded-for', b'')

            if isinstance(ip, list):
                ip = ip[0]

            ip = ip.strip()

            if ip == b'' and self.client is not None:
                self._ip = self.client[0].encode('utf-8')
            else:
                self._ip = ip[:(ip + b',').find(b',')]

        return self._ip

    @property
    def scheme(self):
        if not self._scheme:
            scheme = self.headers.get(b'x-forwarded-proto', b'').strip()

            if scheme == b'' and self.is_secure:
                self._scheme = b'https'
            else:
                self._scheme = scheme or b'http'

        return self._scheme

    @property
    def content_type(self):
        # don't lower() content-type, as it may contain a boundary
        return self.headers.get(b'content-type', b'application/octet-stream')

    @property
    def upgraded(self):
        return self._upgraded

    @upgraded.setter
    def upgraded(self, value):
        self.clear_body()
        self._upgraded = value

    @property
    def has_body(self):
        return (b'content-length' in self.headers or
                b'transfer-encoding' in self.headers)

    def uid(self, length=32):
        if self.client is None:
            port = self.socket.fileno()
        else:
            port = self.client[1]  # 0 - 65535

        prefix = (
            int.to_bytes((port << 16) | (os.getpid() & 0xffff),
                         4, byteorder='big') +
            int(time.time() * 1e7).to_bytes(8, byteorder='big')
        )  # 12 Bytes

        return prefix + os.urandom(max(4, length - len(prefix)))

    def clear_body(self):
        del self._body[:]
        del self._read_buf[:]
        super().clear_body()

    async def body(self, raw=False):
        async for data in self.stream(raw=raw):
            self._body.extend(data)

        return self._body

    def read(self, size=-1):
        return self.recv(size=size, raw=False)

    def eof(self):
        return self._eof and self._read_buf == b''

    async def recv(self, size=-1, raw=True):
        if size == 0 or self.eof():
            return b''

        if self._read_instance is None:
            self._read_instance = self.stream(raw=raw)

        if size == -1:
            return bytes(await self._read_instance.__anext__())

        if len(self._read_buf) < size:
            async for data in self._read_instance:
                self._read_buf.extend(data)

                if len(self._read_buf) >= size:
                    break

        buf = self._read_buf[:size]
        del self._read_buf[:size]
        return bytes(buf)

    async def stream(self, raw=False):
        if self._eof:
            return

        if self.http_continue:
            self.protocol.send_continue()

        if not raw and b'chunked' in self.transfer_encoding:
            buf = bytearray()
            agen = super().recv()
            paused = False
            bytes_unread = 0

            while True:
                if not paused:
                    try:
                        buf.extend(await agen.__anext__())
                    except StopAsyncIteration as exc:
                        if b'\r\n\r\n' not in buf:
                            del buf[:]
                            raise BadRequest(
                                'bad chunked encoding: incomplete read'
                            ) from exc

                if bytes_unread > 0:
                    data = buf[:bytes_unread]

                    yield data
                    del buf[:bytes_unread]

                    bytes_unread -= len(data)

                    if bytes_unread > 0:
                        continue

                    paused = True
                    del buf[:2]
                else:
                    i = buf.find(b'\r\n')

                    if i == -1:
                        if len(buf) > self.protocol.options['buffer_size'] * 4:
                            del buf[:]
                            raise BadRequest(
                                'bad chunked encoding: no chunk size'
                            )

                        paused = False
                        continue

                    try:
                        chunk_size = int(b'0x' + buf[:i].split(b';', 1)[0], 16)
                    except ValueError as exc:
                        del buf[:]
                        raise BadRequest('bad chunked encoding') from exc

                    if chunk_size < 1:
                        break

                    data = buf[i + 2:i + 2 + chunk_size]
                    bytes_unread = chunk_size - len(data)

                    yield data

                    if bytes_unread > 0:
                        paused = False
                        del buf[:]
                    else:
                        paused = True
                        del buf[:chunk_size + i + 4]
        else:
            if (self.content_length >
                    self.protocol.options['client_max_body_size']):
                raise PayloadTooLarge

            async for data in super().recv():
                if -1 < self.content_length < self.body_consumed:
                    # pipelining is not yet supported on a request with a body
                    self.protocol.logger.info('Content-Length mismatch')
                    yield data[:self.content_length - self.body_consumed]
                else:
                    yield data

        self._eof = True

    @property
    def params(self):
        return self.context

    @property
    def query(self):
        try:
            return self.params['query']
        except KeyError:
            self.params['query'] = {}

            if self.query_string != b'':
                self.params['query'] = parse_qs(
                    self.query_string.decode('latin-1'), max_num_fields=100
                )

            return self.params['query']

    @property
    def cookies(self):
        try:
            return self.params['cookies']
        except KeyError:
            self.params['cookies'] = {}

            if b'cookie' in self.headers:
                cookie = self.headers[b'cookie']

                if isinstance(self.headers[b'cookie'], list):
                    cookie = b';'.join(self.headers[b'cookie'])

                for k, v in parse_fields(cookie):
                    k = k.decode('latin-1')

                    if k in self.params['cookies']:
                        self.params['cookies'][k].append(v.decode('latin-1'))
                    else:
                        self.params['cookies'][k] = [v.decode('latin-1')]

            return self.params['cookies']

    async def form(self, max_size=8 * 1048576, max_fields=100):
        try:
            return self.params['post']
        except KeyError as exc:
            self.params['post'] = {}

            if self.content_type.lower().startswith(
                    b'application/x-www-form-urlencoded'):
                async for data in self.stream():
                    self._body.extend(data)

                    if self.body_size > max_size:
                        raise ValueError('form size limit reached') from exc

                if 2 < self.body_size <= max_size:
                    self.params['post'] = parse_qs(
                        self._body.decode('latin-1'),
                        max_num_fields=max_fields
                    )

            return self.params['post']

    async def files(self, max_files=1024, max_file_size=100 * 1048576):
        if self.eof():
            return

        for key, boundary in parse_fields(self.content_type):
            if key == b'boundary' and boundary:
                break
        else:
            raise BadRequest('missing boundary')

        boundary_size = len(boundary)
        header = None
        body = bytearray()

        header_size = 0
        body_size = 0
        content_length = 0
        part = {}  # represents a file received in a multipart request
        paused = False

        if self._read_instance is None:
            self._read_instance = self.stream()

        while max_files > 0 and not self._read_buf.startswith(b'--%s--' %
                                                              boundary):
            data = b''

            if not paused:
                try:
                    data = await self._read_instance.__anext__()
                except StopAsyncIteration as exc:
                    if header_size == 1 or body_size == -1:
                        del body[:]
                        raise BadRequest(
                            'malformed multipart/form-data: incomplete read'
                        ) from exc

            if header is None:
                self._read_buf.extend(data)
                header_size = self._read_buf.find(b'\r\n\r\n') + 2

                if header_size == 1:
                    if len(self._read_buf) > 8192:
                        raise BadRequest(
                            'malformed multipart/form-data: header too large'
                        )

                    paused = False
                else:
                    body.extend(self._read_buf[header_size + 2:])

                    # use find() instead of startswith() to ignore the preamble
                    if self._read_buf.find(b'--%s\r\n' % boundary,
                                           0, header_size) != -1:
                        header = self.header.parse(
                            self._read_buf, header_size=header_size
                        ).headers

                        if b'content-disposition' in header:
                            for k, v in parse_fields(
                                    header[b'content-disposition']):
                                part[k.decode('latin-1')] = v.decode('latin-1')

                        if b'content-length' in header:
                            content_length = int(
                                b'+' + header[b'content-length']
                            )
                            part['length'] = content_length

                        if b'content-type' in header:
                            part['type'] = header[
                                           b'content-type'].decode('latin-1')
                    else:
                        header = {}

                continue

            body.extend(data)
            body_size = body.find(b'\r\n--%s' % boundary, content_length)

            if body_size == -1:
                if len(body) >= max_file_size > boundary_size + 4:
                    sub_part = part.copy()
                    sub_part['data'] = body[:-boundary_size - 4]
                    sub_part['eof'] = False
                    yield sub_part

                    content_length = max(
                        content_length - (len(body) - boundary_size - 4), 0
                    )
                    del body[:-boundary_size - 4]

                paused = False
                continue

            part['data'] = body[:body_size]
            part['eof'] = True
            yield part

            self._read_buf[:] = body[body_size + 2:]
            header = None
            content_length = 0
            part = {}
            paused = True
            max_files -= 1

            del body[:]



# Copyright (c) 2023 nggit

import asyncio
import os
import time

from base64 import urlsafe_b64encode as b64encode
from datetime import datetime, timedelta, timezone
from urllib.parse import quote, unquote_to_bytes

from tremolo.utils import parse_fields
from .http_exceptions import (
    HTTPException,
    BadRequest,
    InternalServerError,
    RangeNotSatisfiable,
    RequestTimeout,
    WebSocketException,
    WebSocketServerClosed
)
from .response import Response
from .websocket import WebSocket

KEEPALIVE_OR_CLOSE = {
    False: b'close',
    True: b'keep-alive'
}
UPGRADE_OR_KEEPALIVE = {
    False: b'keep-alive',
    True: b'upgrade'
}


class HTTPResponse(Response):
    __slots__ = ('request', 'http_chunked', '_headers')

    def __init__(self, request):
        super().__init__(request)

        self.request = request
        self.http_chunked = None

        self._headers = {}

    @property
    def headers(self):
        if self._headers is None:
            raise InternalServerError('headers already sent')

        return self._headers

    def headers_sent(self, sent=False):
        if sent:
            self._headers = None

        return self._headers is None

    def append_header(self, name, value):
        if isinstance(name, str):
            name = name.encode('latin-1')

        if isinstance(value, str):
            value = value.encode('latin-1')

        key = name.lower()

        if key in self.headers:
            self.headers[key].append(name + b': ' + value)
        else:
            self.headers[key] = [name + b': ' + value]

    def set_header(self, name, value=''):
        if isinstance(name, str):
            name = name.encode('latin-1')

        if isinstance(value, str):
            value = value.encode('latin-1')

        if b'\n' in name or b'\n' in value:
            raise InternalServerError

        self.headers[name.lower()] = [name + b': ' + value]

    def set_base_headers(self):
        self.set_header(
            b'Date',
            self.request.protocol.globals.info['server_date']
        )

        if self.request.protocol.globals.info['server_name'] != b'':
            self.set_header(
                b'Server',
                self.request.protocol.globals.info['server_name']
            )

    def set_cookie(self, name, value='', expires=0, path='/', domain=None,
                   secure=False, httponly=False, samesite=None):
        if isinstance(name, str):
            name = name.encode('latin-1')

        value = quote(value).encode('latin-1')
        date_expired = (
            (datetime.now(timezone.utc) + timedelta(seconds=expires))
            .strftime('%a, %d %b %Y %H:%M:%S GMT')
            .encode('latin-1')
        )
        path = quote(path).encode('latin-1')

        cookie = bytearray(
            b'%s=%s; expires=%s; max-age=%d; path=%s' %
            (name, value, date_expired, expires, path)
        )

        for k, v in ((b'domain', domain), (b'samesite', samesite)):
            if v:
                cookie.extend(b'; %s=%s' % (k, quote(v).encode('latin-1')))

        for k, v in ((secure, b'; secure'), (httponly, b'; httponly')):
            if k:
                cookie.extend(v)

        if b'\n' in cookie:
            raise InternalServerError

        self.append_header(b'Set-Cookie', cookie)

    def set_status(self, status=200, message=b'OK'):
        if isinstance(message, str):
            message = message.encode('latin-1')

        if not isinstance(status, int) or b'\n' in message:
            raise InternalServerError

        if b'_line' in self.headers:
            self.headers[b'_line'][1] = b'%d' % status
            self.headers[b'_line'][2] = message
        else:
            self.headers[b'_line'] = [b'HTTP/%s' % self.request.version,
                                      b'%d' % status,
                                      message]

    def get_status(self):
        try:
            _, status, message = self.headers.pop(b'_line')

            return (int(status), message)
        except KeyError:
            return (200, b'OK')

    def set_content_type(self, content_type=b'text/html; charset=utf-8'):
        if isinstance(content_type, str):
            content_type = content_type.encode('latin-1')

        if b'\n' in content_type:
            raise InternalServerError

        self.headers[b'content-type'] = [b'Content-Type: ' + content_type]

    def get_content_type(self):
        try:
            return self.headers.pop(b'content-type')[0][13:].strip(b' \t')
        except KeyError:
            return b'text/html; charset=utf-8'

    def close(self, keepalive=False):
        if not keepalive:
            # this will force the TCP connection to be terminated
            self.request.http_keepalive = False

        super().close()

    async def end(self, data=b'', keepalive=True, **kwargs):
        if self.headers_sent():
            await self.write(data)
        else:
            self.set_base_headers()

            status = self.get_status()
            content_length = len(data)

            if content_length > 0 and (
                        self.request.method == b'HEAD' or
                        status[0] in (204, 205, 304) or 100 <= status[0] < 200
                    ):
                data = b''

            excludes = (b'connection', b'content-length', b'transfer-encoding')

            await self.send(
                b'HTTP/%s %d %s\r\nContent-Type: %s\r\nContent-Length: %d\r\n'
                b'Connection: %s\r\n%s\r\n\r\n%s' %
                (self.request.version,
                 *status,
                 self.get_content_type(),
                 content_length,
                 KEEPALIVE_OR_CLOSE[keepalive and self.request.http_keepalive],
                 b'\r\n'.join(b'\r\n'.join(v) for k, v in self.headers.items()
                              if k not in excludes),
                 data), **kwargs
            )
            self.headers_sent(True)

        self.close(keepalive=keepalive)

    async def write(self, data, chunked=None, buffer_size=16384, **kwargs):
        kwargs['buffer_size'] = buffer_size

        if not self.headers_sent():
            self.set_base_headers()

            if b'connection' not in self.headers:
                # this block is executed when write() is called outside the
                # handler/middleware. e.g. ASGI server
                status = self.get_status()
                no_content = (status[0] in (204, 205, 304) or
                              100 <= status[0] < 200)

                if chunked is None:
                    if self.http_chunked is None:
                        self.http_chunked = (self.request.version == b'1.1' and
                                             not no_content)
                else:
                    self.http_chunked = chunked

                self.set_status(*status)

                if not no_content:
                    self.set_header(b'Content-Type', self.get_content_type())

                if self.http_chunked:
                    self.set_header(b'Transfer-Encoding', b'chunked')

                if self.request.http_keepalive:
                    if status[0] == 101:
                        self.request.upgraded = True
                    elif not (self.http_chunked or
                              b'content-length' in self.headers):
                        # no chunk, no close, no size.
                        # Assume close to signal end
                        self.request.http_keepalive = False

                    self.set_header(
                        b'Connection',
                        UPGRADE_OR_KEEPALIVE[status[0] in (101, 426)]
                    )
                else:
                    self.set_header(b'Connection', b'close')

                if self.request.method == b'HEAD' or no_content:
                    if status[0] not in (101, 426):
                        self.request.http_keepalive = False

                    data = None
                else:
                    self.request.protocol.set_watermarks(
                        high=buffer_size * 4,
                        low=kwargs.get('buffer_min_size', buffer_size // 2)
                    )

            await self.send(
                b' '.join(self.headers.pop(b'_line')) + b'\r\n' +
                b'\r\n'.join(b'\r\n'.join(v) for v in self.headers.values()) +
                b'\r\n\r\n'
            )
            self.headers_sent(True)

        if (self.http_chunked and not self.request.upgraded and
                data is not None):
            await self.send(b'%X\r\n%s\r\n' % (len(data), data), **kwargs)
        else:
            await self.send(data, **kwargs)

    async def sendfile(self, path, offset=0, count=-1, buffer_size=16384,
                       content_type=b'application/octet-stream', executor=None,
                       **kwargs):
        if isinstance(content_type, str):
            content_type = content_type.encode('latin-1')

        kwargs.setdefault(
            'rate', self.request.protocol.options['download_rate']
        )
        kwargs['buffer_size'] = buffer_size
        loop = self.request.protocol.loop

        def run_sync(func, *args):
            if executor is None:
                return loop.run_in_executor(None, func, *args)

            fut = executor.submit(func, *args)

            if isinstance(fut, asyncio.Future):
                return fut

            return asyncio.wrap_future(fut, loop=loop)

        try:
            handle = self.request.protocol.context.RESPONSE_SENDFILE_HANDLE
            await run_sync(handle.seek, offset)  # OSError on a negative offset
        except AttributeError:
            handle = await run_sync(open, path, 'rb')
            self.request.protocol.context.RESPONSE_SENDFILE_HANDLE = handle

            self.request.protocol.add_close_callback(
                self.request.protocol.context.RESPONSE_SENDFILE_HANDLE.close
            )

        st = os.stat(path)
        file_size = st.st_size - offset

        if count > 0:
            file_size = min(count, file_size)

        mdate = time.strftime('%a, %d %b %Y %H:%M:%S GMT',
                              time.gmtime(st.st_mtime)).encode('latin-1')

        if self.request.version == b'1.1' and b'range' in self.request.headers:
            if (b'if-range' in self.request.headers and
                    self.request.headers[b'if-range'] != mdate):
                self.set_content_type(content_type)
                self.set_header(b'Last-Modified', mdate)
                self.set_header(b'Content-Length', b'%d' % file_size)
                self.set_header(b'Accept-Ranges', b'bytes')

                data = True

                while data:
                    data = await run_sync(handle.read, buffer_size)

                    await self.write(data, chunked=False, **kwargs)

                self.close(keepalive=True)
                return

            _range = self.request.headers[b'range']

            if isinstance(_range, list):
                _range = b';'.join(_range)

            ranges = []

            for key, _bytes in parse_fields(_range):
                if not (key == b'bytes' and _bytes):
                    raise BadRequest('bad range')

                try:
                    for v in _bytes.split(b',', 100):
                        v = v.strip()

                        if v.startswith(b'-'):
                            start = file_size + int(v)

                            if start < 0:
                                raise RangeNotSatisfiable

                            ranges.append(
                                (start, file_size - 1, file_size - start)
                            )
                        elif v.endswith(b'-'):
                            start = int(v[:-1])

                            if start >= file_size:
                                raise RangeNotSatisfiable

                            ranges.append(
                                (start, file_size - 1, file_size - start)
                            )
                        else:
                            start, end = v.split(b'-', 1)
                            start = int(start)
                            end = int(end)

                            if end == 0:
                                end = start

                            if start > end or end >= file_size:
                                raise RangeNotSatisfiable

                            ranges.append((start, end, end - start + 1))
                except ValueError as exc:
                    raise BadRequest('bad range') from exc

            self.set_status(206, b'Partial Content')

            if len(ranges) == 1:
                start, end, size = ranges[0]

                self.set_content_type(content_type)
                self.set_header(b'Content-Length', b'%d' % size)
                self.set_header(
                    b'Content-Range', b'bytes %d-%d/%d' %
                    (start, end, file_size)
                )
                await run_sync(handle.seek, start)

                while size > 0:
                    await self.write(
                        await run_sync(handle.read, min(size, buffer_size)),
                        chunked=False, **kwargs
                    )
                    size -= buffer_size
            else:
                boundary = b'----Boundary%s' % b64encode(self.request.uid(24))

                self.set_content_type(
                    b'multipart/byteranges; boundary=%s' % boundary
                )

                for start, end, size in ranges:
                    await self.write(
                        b'--%s\r\nContent-Type: %s\r\n'
                        b'Content-Range: bytes %d-%d/%d\r\n\r\n' %
                        (boundary, content_type, start, end, file_size),
                        **kwargs
                    )
                    await run_sync(handle.seek, start)

                    while size > 0:
                        await self.write(
                            await run_sync(handle.read,
                                           min(size, buffer_size)),
                            **kwargs
                        )
                        size -= buffer_size

                    await self.write(b'\r\n', **kwargs)

                await self.write(b'--%s--\r\n' % boundary, **kwargs)
                await self.write(b'', **kwargs)
        else:
            if (b'if-modified-since' in self.request.headers and
                    self.request.headers[b'if-modified-since'] == mdate):
                self.set_status(304, b'Not Modified')
                await self.write(None)
                return

            self.set_content_type(content_type)
            self.set_header(b'Last-Modified', mdate)
            self.set_header(b'Content-Length', b'%d' % file_size)

            if self.request.version == b'1.1':
                self.set_header(b'Accept-Ranges', b'bytes')

            data = True

            while data:
                data = await run_sync(handle.read, buffer_size)

                await self.write(data, chunked=False, **kwargs)

        self.close(keepalive=True)

    async def handle_exception(self, exc, data=b''):
        if self.request.protocol.is_closing():
            return

        if not isinstance(exc, asyncio.CancelledError):
            self.request.protocol.print_exception(
                exc, quote(unquote_to_bytes(self.request.path))
            )

        # WebSocket
        if isinstance(exc, WebSocketException):
            if isinstance(exc, WebSocketServerClosed):
                data = WebSocket.create_frame(
                    exc.code.to_bytes(2, byteorder='big'), opcode=8
                )
                await self.send(data)

            self.close(keepalive=True)
            return

        # HTTP
        if self.headers_sent():
            self.close()
            return

        if isinstance(exc, TimeoutError):
            exc = RequestTimeout(cause=exc)
        elif not isinstance(exc, HTTPException):
            exc = InternalServerError(cause=exc)

        self.headers.clear()
        self.set_status(exc.code, exc.message)
        self.set_content_type(exc.content_type)

        if not data:
            data = str(exc)

        if isinstance(data, str):
            data = data.encode(exc.encoding)

        await self.end(data, keepalive=False)



# Copyright (c) 2023 nggit

import base64
import hashlib
import os

from .http_exceptions import WebSocketClientClosed, WebSocketServerClosed

_MAGIC = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'


class WebSocket:
    def __init__(self, request, response):
        self.request = request
        self.response = response
        self.protocol = request.protocol

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.request.upgraded:
            await self.accept()

        return await self.receive()

    async def accept(self):
        sha1_hash = hashlib.sha1(  # nosec B303, B324
            self.request.headers[b'sec-websocket-key'] + _MAGIC
        ).digest()
        accept_key = base64.b64encode(sha1_hash)

        self.response.set_status(101, b'Switching Protocols')
        self.response.set_header(b'Upgrade', b'websocket')
        self.response.set_header(b'Sec-WebSocket-Accept', accept_key)
        await self.response.write(None)

    async def recv(self):
        try:
            first_byte, second_byte = await self.request.recv(2)
        except ValueError as exc:
            raise WebSocketClientClosed(
                'connection closed: recv failed'
            ) from exc

        fin = (first_byte & 0x80) >> 7
        opcode = first_byte & 0x0f
        is_masked = (second_byte & 0x80) >> 7
        payload_length = second_byte & 0x7f

        if fin == 0 or opcode == 0:
            raise WebSocketServerClosed(
                'continuation frames are not supported', code=1003
            )

        if payload_length == 126:
            payload_length = int.from_bytes(await self.request.recv(2),
                                            byteorder='big')
        elif payload_length == 127:
            payload_length = int.from_bytes(await self.request.recv(8),
                                            byteorder='big')

            if payload_length > self.protocol.options['ws_max_payload_size']:
                raise WebSocketServerClosed(
                    '%d exceeds maximum payload size (%d)' %
                    (payload_length,
                     self.protocol.options['ws_max_payload_size']),
                    code=1009
                )

        if is_masked:
            masking_key = await self.request.recv(4)
            payload_data = await self.request.recv(payload_length)
            unmasked_data = bytes(
                data_byte ^ masking_key[i % 4] for i, data_byte in enumerate(
                    payload_data)
            )
        else:
            unmasked_data = await self.request.recv(payload_length)

        if opcode == 1:
            return unmasked_data.decode('utf-8')

        if opcode == 2:
            return unmasked_data

        # ping
        if opcode == 9 and payload_length < 126:
            await self.pong(unmasked_data)
            return b''

        # pong
        if opcode == 10 and payload_length < 126:
            return b''

        if opcode == 8:
            code = 1005

            if unmasked_data != b'':
                code = int.from_bytes(unmasked_data[:2], byteorder='big')

            raise WebSocketClientClosed(
                'connection closed (%d)' % code,
                code=code
            )

        raise WebSocketServerClosed(
            'unsupported opcode %x with payload length %d' %
            (opcode, payload_length),
            code=1008
        )

    async def receive(self):
        payload = b''

        while payload == b'':
            # got empty bytes (pong)
            # ping until non-empty bytes received
            timer = self.protocol.loop.call_at(
                self.protocol.loop.time() +
                self.protocol.options['keepalive_timeout'] / 2,
                self._ping
            )

            try:
                payload = await self.recv()
            except TimeoutError as exc:
                raise WebSocketServerClosed('receive timeout',
                                            code=1000) from exc
            finally:
                timer.cancel()

        return payload

    @staticmethod
    def create_frame(payload_data, fin=1, opcode=None, mask=False):
        if opcode is None:
            if isinstance(payload_data, str):
                opcode = 1
            else:
                opcode = 2

        if opcode == 1:
            payload_data = payload_data.encode('utf-8')

        first_byte = (fin << 7) | opcode
        payload_length = len(payload_data)

        if payload_length < 126:
            second_byte = payload_length
            payload_length_data = b''
        elif payload_length < 65536:
            second_byte = 126
            payload_length_data = payload_length.to_bytes(2, byteorder='big')
        else:
            second_byte = 127
            payload_length_data = payload_length.to_bytes(8, byteorder='big')

        if mask:
            second_byte |= (1 << 7)
            masking_key = os.urandom(4)

        frame_header = bytes([first_byte, second_byte]) + payload_length_data

        if mask:
            masked_payload_data = bytes(
                data_byte ^ masking_key[i % 4] for i, data_byte in enumerate(
                    payload_data)
            )
            payload_data = masking_key + masked_payload_data

        return frame_header + payload_data

    async def send(self, payload_data, fin=1, opcode=None):
        await self.response.send(
            __class__.create_frame(payload_data, fin=fin, opcode=opcode)
        )

    def _ping(self):
        # ping only if this connection is still listed,
        # otherwise let the recv timeout drop it
        if self.protocol in self.protocol.globals.connections:
            self.protocol.create_task(self.ping())

    async def ping(self, data=b''):
        await self.send(data, opcode=9)

    async def pong(self, data=b''):
        await self.send(data, opcode=10)

    async def close(self, code=1000):
        await self.send(code.to_bytes(2, byteorder='big'), opcode=8)
        self.response.close(keepalive=True)



# Copyright (c) 2023 nggit

import asyncio

from .exceptions import LifespanError, LifespanProtocolUnsupported


class ASGILifespan:
    def __init__(self, app, **kwargs):
        self._loop = kwargs['loop']
        self._logger = kwargs['logger']

        self._queue = asyncio.Queue()
        self._waiter = self._loop.create_future()
        self._task = self._loop.create_task(self.main(app))

    async def main(self, app):
        try:
            scope = {
                'type': 'lifespan',
                'asgi': {'version': '3.0'}
            }

            await app(scope, self.receive, self.send)
        finally:
            self._waiter.cancel()

    def startup(self):
        self._queue.put_nowait({'type': 'lifespan.startup'})
        self._logger.info('lifespan: startup')

    def shutdown(self):
        self._queue.put_nowait({'type': 'lifespan.shutdown'})
        self._logger.info('lifespan: shutdown')

    async def receive(self):
        data = await self._queue.get()
        self._queue.task_done()

        return data

    async def send(self, data):
        if data['type'] in ('lifespan.startup.complete',
                            'lifespan.shutdown.complete'):
            self._waiter.set_result(None)
            self._logger.info(data['type'])
        elif data['type'] in ('lifespan.startup.failed',
                              'lifespan.shutdown.failed'):
            if 'message' in data:
                message = ': %s' % data['message']
            else:
                message = ''

            raise LifespanError('%s%s' % (data['type'], message))
        else:
            raise LifespanProtocolUnsupported

    async def exception(self, timeout=30):
        timer = self._loop.call_at(self._loop.time() + timeout,
                                   self._waiter.cancel)

        try:
            await self._waiter

            self._waiter = self._loop.create_future()
        except asyncio.CancelledError:
            try:
                exc = self._task.exception()

                if exc:
                    if isinstance(exc, LifespanError):
                        return exc

                    if isinstance(exc, LifespanProtocolUnsupported):
                        self._logger.info(exc)
                    else:
                        self._logger.info(
                            '%s: %s',
                            LifespanProtocolUnsupported.message,
                            str(exc) or repr(exc)
                        )
            except asyncio.InvalidStateError:
                self._logger.warning('lifespan: timeout after %gs', timeout)
        finally:
            timer.cancel()
